package Utils;

import java.awt.IllegalComponentStateException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;

import main.Degree2Path;
import main.Degree2PathState;
import main.Graph;
import main.Node;

public class GraphUtils 
{
	public static int getNonLeafDegree(Node n)
	{
		int leafs = 0;
		for(Node neighbour : n.nonCyclicNeighbours)
			if(neighbour.isLeaf())
				leafs++;
		return n.getNeighbourCount()-leafs;
	}


	public static int countCyclicNeighbours(Node n)
	{
		return n.cyclicNeighbours.size();
	}

	public static void addEdge(Node n1, Node n2)
	{
		if(n1==n2)
			throw new IllegalArgumentException();
		if(n1.isCyclic)
			n2.addCyclicNeighbour(n1);
		else
			n2.addNonCyclicNeighbour(n1);

		if(n2.isCyclic)
			n1.addCyclicNeighbour(n2);
		else
			n1.addNonCyclicNeighbour(n2);
	}

	public static void removeEdge(Node n1, Node n2)
	{
		if(n1.isCyclic)
			n2.removeCyclicNeighbour(n1);
		else
			n2.removeNonCyclicNeighbour(n1);

		if(n2.isCyclic)
			n1.removeCyclicNeighbour(n2);
		else
			n1.removeNonCyclicNeighbour(n2);
	}



	public static void appendDegree2Path(Node connectingNode, Degree2Path mainPath, Degree2Path toAppend)
	{
		//Node[] newPath = new Node[p1.pathNodes.length+p2.pathNodes.length-1];
		if(mainPath.getEnd() != connectingNode)
			Collections.reverse(mainPath.pathNodes);
		if(toAppend.getStart() != connectingNode)
			Collections.reverse(toAppend.pathNodes);

		mainPath.pathNodes.remove(mainPath.pathNodes.size()-1);
		mainPath.pathNodes.addAll(toAppend.pathNodes);
	}

	public static boolean mergeDegree2Paths(ArrayList<Degree2Path> paths, Node connectingNode)
	{
		if(connectingNode.cyclicNeighbours.size() > 2)
			return false;

		Degree2Path p1 = null;
		Degree2Path p2 = null;

		//for(Degree2Path p : paths)
		for(int i = 0; i < paths.size(); i++)
		{
			Degree2Path p = paths.get(i);
			if(p.getEnd() == connectingNode || p.getStart() == connectingNode)
			{
				if(p1== null)
					p1 = p;
				else if(p2 == null)
					p2 = p;
				else 
					throw new IllegalComponentStateException("Node has at most 2 cyclic neighbours but more that 2 connected Degree2Paths");
			}
		}
		if(p2 == null)
			return false;

		//appends p2 to p1 and then removes p2 and moves p1 to the end
		GraphUtils.appendDegree2Path(connectingNode, p1, p2);
		paths.remove(p1);
		p2.state = Degree2PathState.destroyed;
		paths.add(p1);
		return true;
	}

	private static ExpansionRestrictor allowAll = new ExpansionRestrictor() {
		@Override
		public boolean allowExpand(Node current, Node neighbour) {
			return true;
		}
	};

	public static HashSet<Node> expand(Node startNode, ExpandMode mode)
	{
		return expand(startNode, allowAll, mode);
	}
	public static HashSet<Node> expand(Node startNode, final Node notToCross, ExpandMode mode)
	{
		ExpansionRestrictor nonCrossing = new ExpansionRestrictor() {
			@Override
			public boolean allowExpand(Node current, Node neighbour) {
				return neighbour != notToCross;
			}
		};
		return expand(startNode, nonCrossing, mode);
	}
	/*public static HashSet<Node> expand(Node startNode, ExpansionRestrictor restrictor)
	{
		return expand(startNode, null, mode);
	}*/

	public static HashSet<Node> expand(Node startNode, ExpansionRestrictor restrictor, ExpandMode mode)
	{
		//LinkedList<Node> connectedNodes = new LinkedList<Node>();
		HashSet<Node> visited = new HashSet<Node>();

		//now expand from that node
		LinkedList<Node> toExpand = new LinkedList<Node>();
		toExpand.add(startNode);

		while(toExpand.size() > 0)
		{
			Node n = toExpand.poll();
			visited.add(n);
			//connectedNodes.add(n);

			if(mode == ExpandMode.both)
			{
				for(Node neighbour : n.cyclicNeighbours)
				{
					if(!visited.contains(neighbour) && restrictor.allowExpand(n, neighbour))
						toExpand.add(neighbour);
				}
				for(Node neighbour : n.nonCyclicNeighbours)
				{
					if(!visited.contains(neighbour) && restrictor.allowExpand(n, neighbour))
						toExpand.add(neighbour);
				}
			}
			else
			{
				HashSet<Node> neighbours = mode == ExpandMode.cyclic ? n.cyclicNeighbours : n.nonCyclicNeighbours;
				for(Node neighbour : neighbours)
				{
					if(!visited.contains(neighbour) && restrictor.allowExpand(n, neighbour))
						toExpand.add(neighbour);
				}
			}
		}
		return visited;
	}

	/*public static Node findNodeP2(Collection<Node> nodes, boolean withP2)
	{
		for(Node n : nodes)
		{
			if((n.P2.size() > 0) == withP2)
				return n;
		}
		return null;
	}*/

	/**
	 * Determines weather the given connecting component is a caterpillar
	 * To ensure correctness it is important that the given collection is a complete connecting component 
	 * and that the attribute isCyclic is set correctly for each Node.
	 * @param nodes A collection of nodes which should represent a connecting component
	 * @return true if the connecting component is a caterpillar, false otherwise
	 */
	public static boolean isCaterpillar(Collection<Node> nodes)
	{
		for(Node n : nodes)
		{
			if(n.isCyclic || getNonLeafDegree(n) > 2 )
				return false;
		}
		return true;
	}

	/**
	 * Finds the other end of this path. 
	 * @throws IllegalArgumentException if n is neither the start nor the end of the parth
	 * @param path
	 * @param n
	 * @return
	 */
	public static Node getOtherEndNode(Degree2Path path, Node n)
	{
		if(path.getStart() == n)
			return path.getEnd();
		else if(path.getEnd() == n)
			return path.getStart();
		else
			throw new IllegalArgumentException();
	}

	public static Node getOtherNeighbour(Node n, Node notToChoose, ExpandMode mode)
	{
		if(mode == ExpandMode.both || mode == ExpandMode.nonCyclic)
		{
			for(Node neighbour : n.nonCyclicNeighbours)
			{
				if(neighbour != notToChoose)
					return neighbour;
			}
		}
		if(mode == ExpandMode.both || mode == ExpandMode.cyclic)
		{
			for(Node neighbour : n.cyclicNeighbours)
			{
				if(neighbour != notToChoose)
					return neighbour;
			}
		}
		return null;
	}

	public static Node getOtherNeighbour(Node n, HashSet<Node> notToChoose, ExpandMode mode)
	{
		if(mode == ExpandMode.both || mode == ExpandMode.nonCyclic)
		{
			for(Node neighbour : n.nonCyclicNeighbours)
			{
				if(!notToChoose.contains(neighbour))
					return neighbour;
			}
		}
		if(mode == ExpandMode.both || mode == ExpandMode.cyclic)
		{
			for(Node neighbour : n.cyclicNeighbours)
			{
				if(!notToChoose.contains(neighbour))
					return neighbour;
			}
		}
		return null;
	}

	public static String getEdgeName(Node n1, Node n2)
	{
		return n1.name+","+n2.name;
	}
	public static boolean hasEdge(Node n1, Node n2) {
		if(n1.isCyclic)
			return n2.cyclicNeighbours.contains(n1);
		else
			return n2.nonCyclicNeighbours.contains(n1);
	}

	public static Node findLeaf(Collection<Node> candidates)
	{
		for(Node n : candidates)
		{
			if(n.isLeaf())
				return n;
		}
		return null;
	}

	public static boolean hasNonCrossingCyclicPath(Graph g, Node from, Node to, Tuple<Node, Node> edgeNotToCross)
	{
		return hasNonCrossingCyclicPath(g, from, to, edgeNotToCross, null);
	}
	public static boolean hasNonCrossingCyclicPath(Graph g, Node from, Node to, Tuple<Node, Node> edgeNotToCross1, Tuple<Node, Node> edgeNotToCross2)
	{
		//HashSet<Node> visited = new HashSet<Node>();
		boolean[] isVisited = new boolean[g.initialNodesCount];


		LinkedList<Node> toExpand = new LinkedList<Node>();
		toExpand.add(from);

		while(toExpand.size() > 0)
		{
			Node n = toExpand.poll();
			isVisited[n.index] = true;
			//visited.add(n);
			boolean nIsInbannedEdge = edgeNotToCross1.contains(n) || (edgeNotToCross2!=null && edgeNotToCross2.contains(n));
			for(Node neighbour : n.cyclicNeighbours)
			{
				if(nIsInbannedEdge 
						&& edgeNotToCross1.contains(neighbour) || (edgeNotToCross2!=null && edgeNotToCross2.contains(neighbour)))
					continue;
				if(neighbour == to)
					return true;
				//if(!visited.contains(neighbour))
				if(!isVisited[neighbour.index])
					toExpand.add(neighbour);
			}
		}
		return false;
	}


}
