package main;

import java.awt.IllegalComponentStateException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;

import reductionRules.path.PathRedLPYSet;
import visualization.GraphComponent;
import Utils.ExpandMode;
import Utils.GraphUtils;
import Utils.Tuple;

/**
 * represents a normal Degree2Path with at least 3 nodes and one inner path node
 * note: the first and the last node may be the same only if this path is a cycle with this node as the only non-inner-path node
 *		in all other cases there are 2 non-inner-path nodes
 * 
 * @author Steffen Toborg
 */
public class Degree2Path extends Path implements GraphComponent
{
	//Node start;
	//Node end;
	boolean reduced;
	public ArrayList<Node> pathNodes;
	public Token[] tokens;
	public Degree2PathState state;
	public boolean isBridge;

	protected Graph graph;

	/**
	 * Expands the given node as much as possible in order to create a maximum degree-2-path
	 * @throws IllegalArgumentException if the start node is not an inner path node
	 * @param startNode the start node for expanding, this must be an inner path node
	 * @return the new Degree2Path
	 */
	public static Degree2Path getMaximumDegree2Path(Node startNode, Graph g)
	{
		if(!startNode.isCyclic || startNode.cyclicNeighbours.size() != 2)
			throw new IllegalArgumentException();

		Node cyclicNeighbour1 = null;
		Node cyclicNeighbour2 = null;
		for(Node neighbour : startNode.cyclicNeighbours)
		{
			if(cyclicNeighbour1 == null)
				cyclicNeighbour1 = neighbour;
			else
				cyclicNeighbour2 = neighbour;
		}

		LinkedList<Node> pathNodes = new LinkedList<Node>();
		pathNodes.add(startNode);
		expandToDegree2Path_branch(cyclicNeighbour1, startNode, pathNodes, false);
		if(pathNodes.peekFirst() != pathNodes.peekLast())//if not simple circle
		{
			//expand in the other direction, not needed in a simple cycle
			expandToDegree2Path_branch(cyclicNeighbour2, startNode, pathNodes, true);
			return new Degree2Path(new ArrayList<Node>(pathNodes), g);
		}
		else//if simple cycle
		{
			//remove last Node because the last an the first node are the same
			pathNodes.removeLast();
			return new Degree2Circle(new ArrayList<Node>(pathNodes), g);
		}
	}

	private static LinkedList<Node> expandToDegree2Path_branch(Node n, Node prevNode, LinkedList<Node> path, boolean addLast)
	{
		if(addLast)
		{
			path.addLast(n);
			if(n == path.peekFirst())
				return path;//simple cycle found
		}
		else
		{
			path.addFirst(n);
			if(n == path.peekLast())
				return path;//simple cycle found
		}
		if(n.cyclicNeighbours.size() > 2)
			return path;
		for(Node neighbour : n.cyclicNeighbours)
		{
			if(neighbour == prevNode)
				continue;

			return expandToDegree2Path_branch(neighbour, n, path, addLast);
		}
		return path;
	}


	/**
	 * Removes a part of this path inclusive the given nodes.
	 * The nodes will not be removed from the graph and neither will their pendant trees be removed.
	 * @throws IllegalArgumentException If the given nodes are not all inner path nodes
	 * @param first the first node to remove
	 * @param last the last node to remove
	 */
	public void removeNodesFromPath(int first, int last)
	{
		if(last < first || first <= 0 || last >= pathNodes.size()-1)
			throw new IllegalArgumentException();

		Node firstNode = pathNodes.get(first);
		Node firstNode_1 = pathNodes.get(first-1);
		Node lastNode = pathNodes.get(last);
		Node lastNode1 = pathNodes.get(last+1);

		GraphUtils.removeEdge(firstNode_1, firstNode);
		GraphUtils.removeEdge(lastNode, lastNode1);
		GraphUtils.addEdge(firstNode_1, lastNode1);
		

		if(first == last)
			pathNodes.remove(last);
		else{
		Iterator<Node> pathIterator = pathNodes.iterator();
		int count = 0;
		while(pathIterator.hasNext())
		{
			pathIterator.next();
			if(count > last)
				break;
			if(count >= first)
				pathIterator.remove();
			count++;
		}
		}
		if(pathNodes.size() == 4 && getStart() == getEnd())
			this.graph.addThreeCycle(getPathVertex(0), getPathVertex(1), getPathVertex(2));
	}

	/**
	 * Finds all nodes that are contained in the specified part of the path or the nodes's pendant trees 
	 * @param firstNode the first node to remove
	 * @param lastNode the last node to remove
	 * @return A HashSet of the found nodes
	 */
	public HashSet<Node> getPendantTreeNodes(int firstNode, int lastNode)
	{
		HashSet<Node> result = new HashSet<Node>();

		Iterator<Node> pathIterator = pathNodes.iterator();
		int count = 0;
		while(pathIterator.hasNext())
		{
			Node n = pathIterator.next();
			if(count >= firstNode)
			{
				if(count == 0 || count == pathNodes.size()-1)
					result.add(n);
				else
					result.addAll(GraphUtils.expand(n, ExpandMode.nonCyclic));

				if(count >= lastNode)
					break;
			}
			count++;
		}
		return result;
	}

	public Degree2Path(ArrayList<Node> pathNodes, Graph graph) 
	{
		this.pathNodes = pathNodes;
		reduced = false;
		state = Degree2PathState.OK;
		this.graph = graph;
		//isBridge = false;
		//findAndSetTokens();
	}

	public Node getStart() {
		return pathNodes.get(0);
	}

	public Node getEnd() {
		return pathNodes.get(pathNodes.size()-1);
	}

	public boolean isSimpleCircle()
	{
		return false;
	}

	public Node getInnerPathVertex(int vertexNum)
	{
		if( pathNodes.size()-1 > vertexNum && vertexNum > 0) 
			return pathNodes.get(vertexNum);
		else
			return null;
	}

	public Node getPathVertex(int vertexNum)
	{
		return pathNodes.get(vertexNum%pathNodes.size());
	}

	/**
	 * Determines whether this path is chained.
	 * This path is chained if no token separator can be found
	 * @return
	 */
	public boolean isChained()
	{
		return findTokenSeparator() == -1;
	}

	/**
	 * Finds an inner path node that is a token separator. This means there is no Token that contains both of this path's edges around this Node 
	 * @param path
	 * @return the index number of the token separator within this path, this means pathNodes.get(returnVal) is the separator Node, returns -1 if no separator was found
	 */
	public int findTokenSeparator()
	{
		boolean[] isInsideToken = new boolean[pathNodes.size()];

		for(Token t : tokens)
		{
			for(int i = 0; i < t.edges.length-1; i++)
			{
				if(t.edges[i] != -1)
					isInsideToken[t.edges[i]] = true;
			}
		}
		
		if(isSimpleCircle())
		{
			for(int i = 0; i < pathNodes.size(); i++)
			{
				if (!isInsideToken[i])
					return i;
			}
		}
		else
		{
			int[] upperStartToken = getUpperToken(0);
			int[] lowerEndToken = getLowerToken(pathNodes.size()-1);
			
			for(int i = 0; i < upperStartToken.length-1; i++)
			{
				if(upperStartToken[i] != -1)
					isInsideToken[upperStartToken[i]] = true;
			}
			for(int i = 0; i < lowerEndToken.length-1; i++)
			{
				if(lowerEndToken[i] != -1)
					isInsideToken[lowerEndToken[i]] = true;
			}
			
			for(int i = 1; i < pathNodes.size()-1; i++)
			{
				if (!isInsideToken[i])
					return i;
			}
		}
		return -1;
	}


	/**
	 * Determines whether one of the inner path nodes has a Y-graph in it's pendant tree.
	 * For this method to work properly, the property Y must be set correctly for each of the inner path nodes.
	 * @return true if a Y-grahp was found, false otherwise
	 */
	public boolean hasInnerYGraph()
	{
		for(int i = 1; i < pathNodes.size()-1; i++)
		{
			if(PathRedLPYSet.hasYGraph(pathNodes.get(i)))
				return true;
		}
		return false;
	}

	/**
	 * Finds the tokens and saves them in the tokens array.
	 * The old tokens(if there are any) will be overwritten.
	 * Rule: If a node n has one P2 as subgraph then add the token lowerToken(n) U upperToken(n).
	 * 	Else if a node has two P2s as subgraph then and both lowerToken(n) and upperToken(n) separately.
	 * 	Else do nothing.
	 */
	public void findAndSetTokens()
	{
		ArrayList<Token> result = new ArrayList<Token>();
		for(int i = 1; i < pathNodes.size()-1; i++)
		{
			Node n = pathNodes.get(i);
			switch(PathRedLPYSet.getP2Count(n))
			{
			case 0:
				break;
			case 1: 
				result.add(new Token(this, concatArrays(getLowerToken(i), getUpperToken(i)), i));
				break;
			case 2: 
				result.add(new Token(this, getLowerToken(i), i));
				result.add(new Token(this, getUpperToken(i), i));
				break;
			default:
				throw new IllegalComponentStateException();
			}
		}
		tokens = result.toArray(new Token[0]);
	}

	public static int[] concatArrays(int[] array1, int[] array2)
	{
		int[] result = new int[array1.length + array2.length];
		for(int i = 0; i < array1.length; i++)
		{
			result[i] = array1[i];
		}
		for(int i = 0; i < array2.length; i++)
		{
			result[array1.length+i] = array2[i];
		}
		return result;
	}

	protected int[] getUpperToken(int node)
	{
		if(node == pathNodes.size()-1)
			return new int[]{-1};
		else if(node == pathNodes.size()-2)
			return new int[]{pathNodes.size()-1 , -1};
		else
		{
			if(pathNodes.get(node+1).nonCyclicNeighbours.size() == 0)//if crossable
				return new int[]{node+1, node +2};
			else
				return new int[]{node+1};
		}
	}
	
	/**
	 * gets the lower token for this node
	 * @param node the number of the node within this degree-2-path
	 * @return
	 */
	protected int[] getLowerToken(int node)
	{
    //TODO: Gah! Here a token has "nodes" but the Token class has "edges", there
    //are surely off-by-one errors around here...
		if(node == 0)
			return new int[]{-1};
		else if(node == 1)
			return new int[]{-1 , 1};
		else
		{
			if(pathNodes.get(node-1).nonCyclicNeighbours.size() == 0)//if crossable
				return new int[]{node-1, node};
			else
				return new int[]{node};
		}
	}



	@Override
	public boolean contains(Node n) {
		return pathNodes.contains(n);
	}

	@Override
	public Collection<Tuple<Node, Node>> getInnerEdges() {
		ArrayList<Tuple<Node, Node>> result = new ArrayList<Tuple<Node,Node>>();
		for(int i = 0; i < pathNodes.size()-1; i++)
		{
			result.add(new Tuple<Node, Node>(pathNodes.get(i), pathNodes.get(i+1)));
		}
		return result;
	} 
}
