package main;


import graphLoading.IntBasedEdge;

import java.awt.IllegalComponentStateException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;

import java.util.ListIterator;
import java.io.PrintStream;

import solving.GraphCopy;
import visualization.Graph.GraphViewing;
import Utils.BranchingIterator;
import Utils.ExpandMode;
import Utils.GraphUtils;


public class Graph
{

	/**
	 * the number of nodes before applying any rule or algorithm
	 */
	public int initialNodesCount;

	/**
	 * the number of edges before applying any rule or algorithm
	 */
	final public int initialEdgesCount;

	/**
	 * the current set of nodes
	 */
	final public LinkedList<Node> nodes;

	/**
	 * the current state of solving
	 */
	public SolvingState currentState;

	private LinkedList<ThreeCycle> threeCycles;

	/**
	 * Creates a new Graph given the names of the Nodes
	 * The index in the names-array will be the new Node's index
	 * Every edge in the edges-list contains two nodes as Integers which refer to these Node-indexes
	 * @param names
	 * @param edges
	 */
	public Graph(String[] names, ArrayList<IntBasedEdge> edges)
	{
		this.initialNodesCount = names.length;
		this.initialEdgesCount = edges.size();
		this.nodes = new LinkedList<Node>();
		this.threeCycles = new LinkedList<ThreeCycle>();

		Node[] nodeArray = new Node[names.length];
		for(int i = 0; i < names.length; i++)
		{
			nodes.add(nodeArray[i] = new Node(names[i], i));
		}

		for(IntBasedEdge e : edges)
		{
			if(e.left == e.right)
				continue;
			nodeArray[e.left].addNonCyclicNeighbour(nodeArray[e.right]);
			nodeArray[e.right].addNonCyclicNeighbour(nodeArray[e.left]);
		}
		markCyclicNodes();

		validateConsistency();
	}

	/**
	 * This Constructor is for making copies of a graph. The given node list must be complete at this point.
	 * @param nodes
	 * @param initialNodesCount
	 * @param initialEdgesCount
	 */
	public Graph(LinkedList<Node> nodes, int initialNodesCount, int initialEdgesCount)
	{
		this.initialNodesCount = initialNodesCount;
		this.initialEdgesCount = initialEdgesCount;
		this.nodes = nodes;
		this.threeCycles = new LinkedList<ThreeCycle>();
		markCyclicNodes();
	}


	/**
	 * This Constructor is for making copies of a graph. Do not use this constructor for anything else. 
	 * The nodes must be added manually.
	 * Moreover the method markCyclicNodes must be called after setting all the nodes correct.
	 * Beware: Wrong use would cause the whole application to return wrong results.
	 * @param initialNodesCount
	 * @param initialEdgesCount
	 */
	protected Graph(int initialNodesCount, int initialEdgesCount)
	{
		this.initialNodesCount = initialNodesCount;
		this.initialEdgesCount = initialEdgesCount;
		this.threeCycles = new LinkedList<ThreeCycle>();
		nodes = new LinkedList<Node>();
	}

	/**
	 * marks all cyclic nodes as isCyclic
	 * additionally they will be moved from the nonCyclicNeigbbours-lists to the cyclicNeigbbours-lists
	 * this method will most likely throw an exception if any nodes have been deleted yet
	 */
	public void markCyclicNodes()
	{
		int[] nonCyclicNeigbbours = new int[initialNodesCount];
		LinkedList<Node> cyclicCore = new LinkedList<Node>(nodes);
		LinkedList<Node> nonCyclicNodes = new LinkedList<Node>();
		boolean success = false;

		do
		{
			Iterator<Node> iterator = cyclicCore.iterator();
			success = false;
			while(iterator.hasNext())
			{
				Node n = iterator.next();
				if(n.getNeighbourCount() - nonCyclicNeigbbours[n.index] <= 1)
				{
					nonCyclicNodes.add(n);
					iterator.remove();
					for(Node neighbour : n.nonCyclicNeighbours)
						nonCyclicNeigbbours[neighbour.index]++;
					for(Node neighbour : n.cyclicNeighbours)
						nonCyclicNeigbbours[neighbour.index]++;

					success = true;
				}
			}
		}while (success);

		for(Node n : cyclicCore)
			n.setCyclic();
		for(Node n : nonCyclicNodes)
			n.setNonCyclic();
	}

	/**
	 * Finds all degree-2-paths in this graph.
	 * In order to work properly it must be sure that the fields isCyclic and cyclicNeighbours are set correctly for each node.
	 * A degree-2-path has at least one inner path vertex(-> #cyclicneighbours=2). And a start and an end node with #cyclicneighbours>2.
	 * Those paths which don't have a start or an end with #cyclicneighbours>2 must be simple circes without connection to another cyclic node and are added as degree-2-circles(a subclass of degree-2-path).
	 * @return The list of degree-2-paths
	 */
	/*public ArrayList<Degree2Path> findDegree2Paths()
	{
		return findDegree2Paths(nodes);
	}*/

	/**
	 * Finds all degree-2-paths in this graph within the given set of allowed nodes.
	 * Node: If only one inner path node of a degree-2-path is in this set the path will be found and added to the result list.
	 * 	But for this method to work correctly the list should be complete. This means of every degree-2-path that can be found there should be either all path nodes (not only the inner path nodes) in this set or none. 
	 * In order to work properly it must be sure that the fields isCyclic and cyclicNeighbours are set correctly for each node.
	 * A degree-2-path has at least one inner path vertex(-> #cyclicneighbours=2). And a start and an end node with #cyclicneighbours>2.
	 * Those paths which don't have a start or an end with #cyclicneighbours>2 must be simple circes without connection to another cyclic node and are added as degree-2-circles(a subclass of degree-2-path).
	 * @return The list of degree-2-paths
	 * @param nodesToSearchWith The set of nodes for the search
	 */
	/*public ArrayList<Degree2Path> findDegree2Paths(Collection<Node> nodesToSearchWith)
	{
		ArrayList<Degree2Path> pathList = new ArrayList<Degree2Path>();

		for(Node n : nodesToSearchWith)
			n.degree2Paths.clear();

		MarkingIterator<Node> nodeIterator = new MarkingIterator<Node>(nodesToSearchWith);
		while(nodeIterator.hasNext())
		{
			Node n = nodeIterator.next();
			if(n.isCyclic && n.cyclicNeighbours.size()==2)
			{
				Degree2Path newPath = Degree2Path.getMaximumDegree2Path(n, this);
				nodeIterator.markAll(newPath.pathNodes);
				if(!newPath.isSimpleCircle())//not needed for simple cycles
				{
					newPath.getStart().degree2Paths.add(newPath);
					if(newPath.getStart() != newPath.getEnd())
						newPath.getEnd().degree2Paths.add(newPath);
				}
				pathList.add(newPath);
			}
		}
		return pathList;
	}*/

	/*public void addTokenGenerator(Node n)
	{
		tokenGenerators.add(n);
	}*/

	public void addThreeCycle(Node n1, Node n2, Node n3)
	{
		threeCycles.add(new ThreeCycle(n1, n2, n3));
	}

	public BranchingIterator<ThreeCycle> getThreeCycleBranchingIterator()
	{
		return new BranchingIterator<ThreeCycle>(threeCycles);
	}

	public boolean hasThreeCycles()
	{
		return threeCycles.size() > 0;
	}

	public void findAndSetThreeCycles()
	{//n1<n2<n3
		for(Node n1 : nodes)
		{
			for(Node n2 : n1.cyclicNeighbours) if(n1.index < n2.index)
			{
				for(Node n3 : n1.cyclicNeighbours) if(n2.index < n3.index)
				{
					if(n2.cyclicNeighbours.contains(n3))
						addThreeCycle(n1, n2, n3);
				}
			}
		}
	}

	private boolean hasP2s;
	public void onP2Found()
	{
		hasP2s = true;
	}
	public boolean hasP2s()
	{
		return hasP2s;
	}

	public void validateConsistency()
	{
		if(!Start.debugMode) return;
		
		GraphCopy copy = new GraphCopy(this);
		copy.markCyclicNodes();
		for(Node ng : this.nodes)
		{
			Node nCopy = copy.getNodeClone(ng);
			if(ng.isCyclic != nCopy.isCyclic)
				{
				GraphViewing.showGraphAndWaitForClose(this, "Cyclic not set correctly", ng);
				throw new IllegalComponentStateException();
				}
		}
		
		if(!nodes.isEmpty())
		{
			HashSet<Node> expandedSet = GraphUtils.expand(nodes.getFirst(), ExpandMode.both);
			if(expandedSet.size()>nodes.size())
				throw new IllegalComponentStateException();
			for(Node n : expandedSet)
			{
				if(n.deleted)
					throw new IllegalComponentStateException();
			}
		}
		for(Node n : nodes)
		{
			if(n.deleted)
				throw new IllegalComponentStateException();
			if(n.isCyclic)
			{
				for(Node neighbour : n.nonCyclicNeighbours)
					if(!neighbour.cyclicNeighbours.contains(n))
						throw new IllegalComponentStateException();

				for(Node neighbour : n.cyclicNeighbours)
					if(!neighbour.cyclicNeighbours.contains(n))
						throw new IllegalComponentStateException();
			}
			else
			{
				for(Node neighbour : n.nonCyclicNeighbours)
					if(!neighbour.nonCyclicNeighbours.contains(n))
						throw new IllegalComponentStateException();

				for(Node neighbour : n.cyclicNeighbours)
					if(!neighbour.nonCyclicNeighbours.contains(n))
						throw new IllegalComponentStateException();
			}

			//Node neighbour = GraphUtils.getOtherNeighbour(n, (Node)null, ExpandMode.cyclic);
			//boolean shouldBeCyclic = neighbour!=null && null!=BSS36_Claws.getNonCrossingPathsLastNode(neighbour, n, new Tuple<Node, Node>(neighbour, n));
			boolean shouldBeCyclic = n.cyclicNeighbours.size() >= 2;
			if(shouldBeCyclic != n.isCyclic)
			{
				GraphViewing.showGraphAndWaitForClose(this, "");
				throw new IllegalComponentStateException();
			}
		}
	}

  public void print(PrintStream out){
    ListIterator<Node> li = nodes.listIterator();
    Iterator<Node> hi;
    Node v, w;

    while(li.hasNext()){
      v = li.next();

      hi = v.cyclicNeighbours.iterator();
      while(hi.hasNext()){
        w = hi.next();
        if(v.index < w.index) out.println(v.name + " " + w.name);
      }
      hi = v.nonCyclicNeighbours.iterator();
      while(hi.hasNext()){
        w = hi.next();
        if(v.index < w.index) out.println(v.name + " " + w.name);
      }
    }
  }

}
