package main;
import java.awt.IllegalComponentStateException;
import java.util.HashSet;
import java.util.LinkedList;

import visualization.GraphComponent;

/**
 * A normale Node of a graph
 * @author Steffen Toborg
 */
public class Node implements GraphComponent
{
	public int index;
	public String name;
	public boolean isCyclic;
	public boolean deleted;
	public boolean consistent;
//	public LinkedList<Node> L;//leaves as neighbours
//	public LinkedList<Node> P2;
//	public LinkedList<Node> Y;//Y-Graphs
	//public LinkedList<Degree2Path> degree2Paths;

	/**
	 * A set of all neighbours that are contained in a cycle or in a bridge between 2 cycles.
	 */
	public HashSet<Node> cyclicNeighbours;
	
	/**
	 * A set of all neighbours that are not contained in a cycle or in a bridge between 2 cycles.
	 */
	public HashSet<Node> nonCyclicNeighbours;



	public Node(String name, int index)  {
//		this.L  = new LinkedList<Node>();
//		this.P2 = new LinkedList<Node>();
//		this.Y = new LinkedList<Node>();
//		this.degree2Paths = new LinkedList<Degree2Path>();
		this.cyclicNeighbours = new HashSet<Node>();
		this.nonCyclicNeighbours = new HashSet<Node>();
		this.isCyclic = false;
		this.deleted = false;
		this.name = name;
		this.index  = index;
		this.consistent = true;
	}

	@Override
	public int hashCode() {
		return index;
	}
	
	public void setConsistent(boolean value)
	{
		consistent = value;
	}
	public boolean isConsistent()
	{
		return consistent;
	}

	public void setCyclic()
	{
		if(deleted)
			throw new IllegalComponentStateException();
		isCyclic = true;
		for(Node neighbour : nonCyclicNeighbours)
			neighbour.onNeighbourMarkedCyclic(this);
		for(Node neighbour : cyclicNeighbours)
			neighbour.onNeighbourMarkedCyclic(this);
	}

	public void setNonCyclic()
	{
		if(deleted)
			throw new IllegalComponentStateException();
		isCyclic = false;
		for(Node neighbour : nonCyclicNeighbours)
			neighbour.onNeighbourMarkedNonCyclic(this);
		for(Node neighbour : cyclicNeighbours)
			neighbour.onNeighbourMarkedNonCyclic(this);
	}

	/**
	 * Called if a neighbour of this Node is marked cyclic
	 * @param n The neighbour that is now marked cyclic
	 */
	private void onNeighbourMarkedCyclic(Node n)
	{
		nonCyclicNeighbours.remove(n);
		cyclicNeighbours.add(n);
	}

	/**
	 * Called if a neighbour of this Node is marked non-cyclic
	 * @param n The neighbour that is now marked non-cyclic
	 */
	private void onNeighbourMarkedNonCyclic(Node n)
	{
		cyclicNeighbours.remove(n);
		nonCyclicNeighbours.add(n);
	}

	public void addNonCyclicNeighbour(Node node){
		if(node==null)
			throw new IllegalArgumentException();
		nonCyclicNeighbours.add(node);
	}

	public void addCyclicNeighbour(Node node){
		cyclicNeighbours.add(node);
	}

	public boolean removeNonCyclicNeighbour(Node node){
		return nonCyclicNeighbours.remove(node);
	}

	public boolean removeCyclicNeighbour(Node node){
		return cyclicNeighbours.remove(node);
	}


	public int getNeighbourCount()
	{
		return nonCyclicNeighbours.size()+cyclicNeighbours.size();
	}

	public boolean isLeaf()
	{
		return getNeighbourCount() <= 1;
	}

	public void addCustomName(String name, Node otherNode)
	{

	}

	public void removeFromNeighbourLists() 
	{
		if(isCyclic)
		{
			for(Node n : nonCyclicNeighbours)
				n.removeCyclicNeighbour(this);
			for(Node n : cyclicNeighbours)
				n.removeCyclicNeighbour(this);
		}
		else
		{
			for(Node n : nonCyclicNeighbours)
				n.removeNonCyclicNeighbour(this);
			for(Node n : cyclicNeighbours)
				n.removeNonCyclicNeighbour(this);
		}
	}
	@Override
	public String toString() {

		return name+"("+index+")";
	}
	
	public boolean isSingleton()
	{
		return getNeighbourCount() == 2;
		//return nonCyclicNeighbours.size() == 0;
	}
	/*public void clearLPY()
	{
		L.clear();
		P2.clear();
		Y.clear();
	}*/

//	/**
//	 * recomputes the L, P2 and Y lists of this node
//	 * Note that it is intended that some neighbours don't have their lists set 
//	 * as this method is supposed to be called while branching through every pendant-tree of every cyclic node recursively
//	 */
	/*public void computeLPY()
	{
		for(Node neighbour : nonCyclicNeighbours)
		{
			if(neighbour.isLeaf())
				L.add(neighbour);

			if(neighbour.L.size() == 1)
				P2.add(neighbour);

			if(neighbour.P2.size() >=  2 )
				Y.add(neighbour);
		}
	}*/

	@Override
	public boolean contains(Node n) {
		return n == this;
	}
}
