package reductionRules;

import java.awt.IllegalComponentStateException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;

import main.Degree2Circle;
import main.Degree2Path;
import main.Degree2PathState;
import main.Graph;
import main.Node;
import reductionRules.path.PRR1_CutOffYGraph;
import reductionRules.path.PRR2_CutOutYGraph;
import reductionRules.path.PRR3_ContractPath;
import reductionRules.path.PRR4_SplitPath;
import reductionRules.path.PRR5_ReduceChainedPath;
import reductionRules.path.PathReductionRule;
import reductionRules.tree.TRR1_ReduceLeaves;
import reductionRules.tree.TRR2_CutOffYGraph;
import reductionRules.tree.TRR3_Reduce2Claws;
import reductionRules.tree.TRR4_RemoveEndPoint;
import reductionRules.tree.TRR5_RemoveEndYGraph;
import reductionRules.tree.TreeRedLPYSet;
import reductionRules.tree.TreeReductionRule;
import result.ResultList;
import solving.GraphCopy;
import visualization.Graph.GraphViewing;
import Utils.ConnCompIterator;
import Utils.GraphUtils;

public class Reduction {
	public static TreeReductionRule[] treeReductionRules = new TreeReductionRule[]{
		new TRR1_ReduceLeaves(), 
		new TRR2_CutOffYGraph(), 
		new TRR3_Reduce2Claws(), 
		new TRR4_RemoveEndPoint(), 
		new TRR5_RemoveEndYGraph()
	};
	public static PathReductionRule[] pathRulesYGraphs = new PathReductionRule[]{ new PRR1_CutOffYGraph(), new PRR2_CutOutYGraph() };
	public static PathReductionRule[] pathRulesTokens = new PathReductionRule[]{ new PRR3_ContractPath(), new PRR4_SplitPath(), new PRR5_ReduceChainedPath()};


	/**
	 * Finds Degree2Path's and applies the path-reduction-rules to them until no more Degree2Path can be found
	 * @return the value by that k has been reduced due to these rules
	 */
	public static int applyPathReductionRules(Graph g, ResultList resultList)
	{
		int result = 0;
		HashSet<Node> visited = new HashSet<Node>();
		while(true)
		{
			boolean success = false;
			for(Node n : g.nodes)
			{
				if(n.isCyclic && n.cyclicNeighbours.size()==2 && !visited.contains(n))
				{
					Degree2Path newPath = Degree2Path.getMaximumDegree2Path(n, g);
					result += applyPathReductionRules(g, newPath, resultList);
					g.validateConsistency();
					if(newPath.state == Degree2PathState.destroyed)
					{
						visited.remove(newPath.getEnd());
						visited.remove(newPath.getStart());
					}
					else
						visited.addAll(newPath.pathNodes);
					success = true;
					break;
				}
			}
			if(!success)
				break;
		}
		return result;
	}


	public static int applyPathReductionRules(Graph g, Degree2Path path, ResultList resultList)
	{
		int kReduced = 0;
		for(PathReductionRule rule : pathRulesYGraphs)
		{
			g.validateConsistency();
			kReduced += rule.apply(g, path, resultList);
			g.validateConsistency();
		}

		if(!path.hasInnerYGraph())
		{
			path.findAndSetTokens();
			for(PathReductionRule rule : pathRulesTokens)
			{
				kReduced += rule.apply(g, path, resultList);

				if(path.state == Degree2PathState.destroyed)
					break;
			}
		}
		return kReduced;
	}

	/**
	 * Applies the tree reduction rules to all cyclic nodes.
	 * From each cyclic node the tree reduction rules will be applied recursively to all nodes within it's pendant tree.
	 * @return the reduction of k due to these rules
	 */
	public static int applyTreeReductionRules(Graph g, ResultList resultList)
	{
		g.validateConsistency();
		if(g.nodes.isEmpty())
			return 0;

		int kReduced = 0;

		ConnCompIterator compIterator = new ConnCompIterator(g);
		while(compIterator.hasNext())
		{
			HashSet<Node> componentNodes = compIterator.next();
			LinkedList<Node> cyclicNodes = new LinkedList<Node>();
			for(Node n : componentNodes)
			{
				if(n.isCyclic)
					cyclicNodes.add(n);
			}

			if(cyclicNodes.size()== 0)
			{
				Node minNbrNode = null;
				Node leaf = GraphUtils.findLeaf(componentNodes);

				kReduced += applyTreeReductionRules(g, leaf, null, resultList);

//				System.out.println(leaf.name);
				if(GraphUtils.isCaterpillar(componentNodes))
					g.nodes.removeAll(componentNodes);
			}
			else
			{
				for(Node n : cyclicNodes)
				{
					g.validateConsistency();
					kReduced += applyTreeReductionRules(g, n, null, resultList);
					g.validateConsistency();
				}
			}
			//		return result;
		}
		return kReduced;
	}


	/**
	 * Applies the Tree Reduction Rules recursively beginning with the given node 
	 * @param n the current node to apply the rules to
	 * @param parent The parent of n in the branch tree(to avoid going backwards)
	 * @return the reduction of k due to these rules
	 */
	public static int applyTreeReductionRules(Graph g, Node n, Node parent, ResultList resultList)
	{
		return applyTreeReductionRules(g, n, new TreeRedLPYSet(n), parent, resultList);
	}

	/**
	 * Applies the Tree Reduction Rules recursively beginning with the given node 
	 * @param n the current node to apply the rules to
	 * @param thisLPY the LPY set for n, values have to be set before reduction and before return
	 * @param parent The parent of n in the branch tree(to avoid going backwards)
	 * @return the reduction of k due to these rules
	 */
	public static int applyTreeReductionRules(Graph g, Node n, TreeRedLPYSet thisLPY, Node parent, ResultList resultList)
	{
		//n.clearLPY();
		g.validateConsistency();
		int returnVal = 0;
		Iterator<Node> neighbourIterator = n.nonCyclicNeighbours.iterator();
		while(neighbourIterator.hasNext())
		{
			Node neighbour = neighbourIterator.next();
			if(neighbour != parent)
			{
				if(neighbour.deleted)
					throw new IllegalComponentStateException();

				TreeRedLPYSet lpyNeighbour = new TreeRedLPYSet(neighbour);
				returnVal += applyTreeReductionRules(g, neighbour, lpyNeighbour, n, resultList);

				if(neighbour.deleted)
					neighbourIterator.remove();
				else
				{
					if(neighbour.isLeaf())
						thisLPY.L.add(neighbour);
					else if(lpyNeighbour.P2.size() >=  2 )
						thisLPY.Y.add(neighbour);
					else if(lpyNeighbour.L.size() == 1)
						thisLPY.P2.add(neighbour);
				}
			}
		}

		for(TreeReductionRule rule : treeReductionRules)
			returnVal += rule.apply(g, n, thisLPY, resultList);

		return returnVal;
	}

	/**
	 * Finds connected components in this graphs and deletes every one of them that is a caterpillar.
	 */
	public static void removeCaterpillars(Graph g) 
	{
		Iterator<HashSet<Node>> connCompIterator = new ConnCompIterator(g);
		while(connCompIterator.hasNext())
		{
			HashSet<Node> comp = connCompIterator.next();
			if(GraphUtils.isCaterpillar(comp))
				g.nodes.removeAll(comp);
		}
	}
}
