package reductionRules.path;

import java.util.ArrayList;

import main.Degree2Path;
import main.Degree2PathState;
import main.Graph;
import main.Node;
import reductionRules.Reduction;
import result.ResultList;
import solving.newSolver.SolvingUtils;

/**
 * Path Reduction Rule 4:
 * Let P denote a degree-2-path with K(P)!={} (at least one token) and P is not chained. Let v(i) be a token separator of P.
 * Then repalce it's pendant tree T^v(i) by two copies of T^v(i), connect one to v(i-1) (by inserting an edge between its connection point and v(i-1)), and connect the other to v(i+1)
 * Note: In this implementation the original will not be removed, instead it serves as the first copy as it is already connected to v(i-1).
 * @author Steffen Toborg
 */
public class PRR4_SplitPath extends PathReductionRule 
{
	@Override
	protected int applyRule(Graph g, Degree2Path path, ResultList resultList) 
	{
		g.validateConsistency();
		int kReduced = 0;
		if(path.tokens.length > 0)
		{

			int separator = path.findTokenSeparator();

			if(separator == -1)//if chained
				return 0;

			//if(path.isSimpleCircle())
			//	System.out.println("");

			Node separatorNode = path.getPathVertex(separator);
			Node separatorNextNode = path.getPathVertex(separator+1);

			deleteEdge(separatorNode, separatorNextNode);

			//make a copy of the separator and it's pendant tree and connect it with the next path node
			Node pendantTreeCopy = makePendantTreeCopy(g, separatorNode, null);
			addEdge(pendantTreeCopy, separatorNextNode);

			onDestroyPath(path);

			//TODO: If I read correctly, Degree2Path.isSimpleCircle() is always false
			if(path.isSimpleCircle())
			{
				path.pathNodes.get(0).setNonCyclic();
				path.pathNodes.get(path.pathNodes.size()-1).setNonCyclic();
				g.validateConsistency();
				//kReduced += Reduction.applyTreeReductionRules(g, GraphUtils.findLeaf(g.nodes), null, resultList); // we have new deleted edges here
				kReduced += Reduction.applyTreeReductionRules(g, separatorNode, null, resultList); // we have new deleted edges here
			}
			else
			{
				Node startNode = path.getStart();
				Node endNode = path.getEnd();

				ArrayList<Node> hooks = new ArrayList<Node>(2);
				if(startNode==endNode)
					hooks.add(SolvingUtils.validateCyclicAndFindHook(startNode, null));
				else
				{
					hooks.add(SolvingUtils.validateCyclicAndFindHook(startNode, null));
					hooks.add(SolvingUtils.validateCyclicAndFindHook(endNode, null));
				}

				kReduced += SolvingUtils.applyReductionForHooks(g, hooks, resultList);

				//SolvingUtils.findHookAndApplyReduction(g, startNode, null, resultList);
				//if(startNode != endNode)
				//	SolvingUtils.findHookAndApplyReduction(g, endNode, null, resultList);

			}
		}
		g.validateConsistency();
		return kReduced;
	}

	private void onDestroyPath(Degree2Path path)
	{
		for(int i = 1; i < path.pathNodes.size()-1; i++)
		{
			path.pathNodes.get(i).setNonCyclic();
		}
		//path.getStart().degree2Paths.remove(path);
		//if(path.getStart() != path.getEnd())//if startnode == endNode this path is allready removed
		//	path.getEnd().degree2Paths.remove(path);

		path.state = Degree2PathState.destroyed;
	}
	//
	//	private Node findLeaf(Degree2Path path)
	//	{
	//		for(Node n : path.pathNodes)
	//		{
	//			for(Node neighbour : n.nonCyclicNeighbours)
	//			{
	//				if(neighbour.getNeighbourCount() == 1)
	//					return neighbour;
	//				else if(neighbour.getNeighbourCount() == 2)
	//					return GraphUtils.getOtherNeighbour(neighbour, n, ExpandMode.nonCyclic);
	//				else if(neighbour.getNeighbourCount() == 3)
	//				{
	//					Node p2Node = GraphUtils.getOtherNeighbour(neighbour, n, ExpandMode.nonCyclic);
	//					return GraphUtils.getOtherNeighbour(p2Node, neighbour, ExpandMode.nonCyclic);
	//				}
	//			}
	//
	//		}
	//		return null;
	//	}



	/**
	 * makes a copy of this node and it's pendant tree and adds the copies to the node-list of the graph
	 * this method recursively calls itself until every node in the original node's branching tree is visited
	 * if of the nodes contains a cyclic node in it's nonCyclicNeighbours set this call will result in an endless loop 
	 * @param g the Graph to ad the created nodes to
	 * @param toCopy the Node to copy
	 * @param prevNode the previous node, this is to prevent going backwards in the branching tree
	 * @return
	 */
	private Node makePendantTreeCopy(Graph g, Node toCopy, Node prevNode)
	{
		Node returnNode = new Node(toCopy.name, g.initialNodesCount++);
		onNodeAdded(returnNode);
		g.nodes.add(returnNode);

		for(Node neighbour : toCopy.nonCyclicNeighbours)
		{
			if(neighbour != prevNode)
				addEdge(returnNode, makePendantTreeCopy(g, neighbour, toCopy));
			//returnNode.addNonCyclicNeighbour(makePendantTreeCopy(g, neighbour, toCopy));
		}
		return returnNode;
	}

	@Override
	public String getName() {
		return "Path-Reduction-Rule 4: Split path";
	}
}
