package reductionRules.path;

import java.util.ArrayList;

import result.ResultList;

import visualization.Graph.GraphViewing;

import com.ctc.wstx.dtd.TokenContentSpec;

import Utils.GraphUtils;
import Utils.Tuple;

import main.Degree2Path;
import main.Degree2PathState;
import main.Graph;
import main.Node;
import main.Token;

/**
 * Path Reduction Rule 5:
 * Let P be a chained degree-2-path with |K(P)|>=3 (at least 3 tokens). 
 * Let M_P:={i in N | K(v(i)) != {}} (the collection of the indexes of all token creating nodes), let p:=min(M_P) and let q:=max(M_P). 
 * If |K(P)| is even, then delete T_p+1,q-1, insert the edge {v(p),v(q)} and reduce k by (|K(P)|-2)/2;
 * otherwise delete T_p+1,q, insert the edge {v(p),v(q+1)} and reduce k by (|K(P)|-1)/2.
 * @author Steffen Toborg
 */
public class PRR5_ReduceChainedPath extends PathReductionRule
{
	@Override
	protected int applyRule(Graph g, Degree2Path path, ResultList resultList) {
		if(path.tokens.length >= 3 && path.isChained())
		{
			int p = path.tokens[0].generator;
			int q = path.tokens[path.tokens.length-1].generator;
			int tokensCount = path.tokens.length;
			boolean isEven = tokensCount % 2 == 0;

			//without this the path would reduced to a cyclic 2-nodes path with a double edge (start - inner - end, start = end) 
			if(path.getStart() == path.getEnd() && !isEven && p <=1 && q >= path.pathNodes.size()-2) 
			{
				if(path.tokens.length == 3)
					return 0;
				q = path.tokens[path.tokens.length-2].generator;
				tokensCount--;
				isEven = !isEven;
			}


			int returnVal = 0;

			boolean eDelEdge = true;
			ArrayList<Tuple<Node, Node>> eDelEdges = new ArrayList<Tuple<Node,Node>>(tokensCount/2 +1);
			ArrayList<Tuple<Node, Node>> eNotDelEdges = new ArrayList<Tuple<Node,Node>>(tokensCount/2);

			for (int i = 0; i < (isEven ? tokensCount-1 : tokensCount); i++) {
				Token t = path.tokens[i];

				//get last edge of token that is not -1
				int e = t.edges[t.edges.length-1]==-1 ? t.edges[t.edges.length-2] : t.edges[t.edges.length-1];
				Node n1 = path.getPathVertex(e-1);
				Node n2 = path.getPathVertex(e);
				if(eDelEdge)
					eDelEdges.add(new Tuple<Node, Node>(n1, n2));	
				else
					eNotDelEdges.add(new Tuple<Node, Node>(n1, n2));	

				eDelEdge = !eDelEdge;
			}

			if(path.isSimpleCircle())
			{
				for (Tuple<Node, Node> tuple : eDelEdges) {
					resultList.add(true, tuple.left, tuple.right);
					GraphUtils.removeEdge(tuple.left, tuple.right);
				}

				for(int i = 0; i < path.pathNodes.size(); i++)
				{
					path.pathNodes.get(i).setNonCyclic();
				}
				path.state = Degree2PathState.destroyed;

				return eDelEdges.size();
			}
			//if even
			if(isEven)
			{
				resultList.addEdgeReplacer(true, path.getPathVertex(p), path.getPathVertex(q), eDelEdges, eNotDelEdges);
				removeSubPath(g, path, p+1, q-1);
				returnVal = (tokensCount-2)/2;
			}
			//if uneven
			else
			{
				Node q1Node = path.getPathVertex(q+1);
				if(GraphUtils.getNonLeafDegree(q1Node) == 2)//add a second conditionNode
					resultList.addEdgeReplacer(true, path.getPathVertex(p), q1Node, q1Node, path.getPathVertex(q+2), eDelEdges, eNotDelEdges);
				else
					resultList.addEdgeReplacer(true, path.getPathVertex(p), q1Node, eDelEdges, eNotDelEdges);
				removeSubPath(g, path, p+1, q);
				returnVal = (tokensCount-1)/2;
			}
			path.findAndSetTokens();//reset tokens

			//GraphViewing.showGraphAndWaitForClose(g, "PRR5");
			return returnVal;
		}
		return 0;
	}

	/*private Pair<Integer, Integer> findCirclePQ(Degree2Path path)
	{
		Token lastToken = null;
		Token firstToken = null;
		Pair<Integer, Integer> returnVal = null;

		int lastEdge = path.tokens[0].edges[path.tokens[0].edges.length-1];
		//assume that Tokens are sorted by min Node number
		for(int i = 1; i < path.tokens.length+2; i++)
		{
			Token t = path.tokens[i % path.tokens.length];
			if(t.edges[0] <= lastEdge) //tokens have common edges
			{
				lastToken = t;
			}
			else if(firstToken == null)//no common edges
			{
				firstToken = t;
				returnVal = new Pair<Integer, Integer>(firstToken.generator, lastToken.generator);
			}
			else // no common edges but first token allready set
				return null;

			lastEdge = t.edges[t.edges.length-1];
		}
		if(firstToken == null)
		{
			int p = path.tokens[0].generator;
			int q = path.tokens[path.tokens.length-1].generator;
			return new Pair<Integer, Integer>(p, q);
		}
		else
			return new Pair<Integer, Integer>(firstToken.generator, lastToken.generator);
	}*/

	/*private void deleteSubTree(Graph g, Degree2Path path, int firstNode, int lastNode)
	{
		GraphUtils.removeEdge(path.pathNodes.get(firstNode-1), path.pathNodes.get(firstNode));
		GraphUtils.removeEdge(path.pathNodes.get(lastNode), path.pathNodes.get(lastNode+1));

		//assume there are no Y-Graphs
		for(int i = firstNode; i <= lastNode; i++)
		{
			Node n = path.pathNodes.get(i);
			g.removeNode(n);
			for(Node p2 : n.P2)
				g.nodes.remove(p2.L.peek());
			g.nodes.removeAll(n.nonCyclicNeighbours);
		}
	}*/


	@Override
	public String getName() {
		return "Path-Reduction-Rule 5: Reduce chained path";
	}
}

